/*
 *        HttXimpUtil.c 1.4 91/09/01  
 * 
 * Derived from XimpUtil.c	2.1 91/08/13 11:28:46 FUJITSU LIMITED. */
/*
****************************************************************************

              Copyright 1991, by FUJITSU LIMITED
              Copyright 1991, by Sun Microsystems, Inc.	      

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of FUJITSU LIMITED and Sun
Microsystems, Inc. not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.  FUJITSU LIMITED and Sun Microsystems, Inc. makes no
representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.

FUJITSU LIMITED AND SUN MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL FUJITSU LIMITED AND SUN
MICROSYSTEMS, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

Auther: Takashi Fujiwara	FUJITSU LIMITED
				fujiwara@a80.tech.yk.fujitsu.co.jp
	Kiyoaki oya		FUJITSU LIMITED
	Riki	Kawaguchi	FUJITSU LIMITED
	Tomoki	Kuriki		FUJITSU LIMITED
	Hideki  Hiura           (hhiura@Sun.COM) Sun Microsystems, Inc.

****************************************************************************
*/

#ifndef lint
#ident	"@(#)HttXimpUtil.c	1.23	93/02/21 SMI"
static char     derived_from_sccs_id[] = "XimpUtil.c	2.1 91/08/13 11:28:46 FUJITSU LIMITED.";
#endif

#include <stdio.h>
#include "Htt.h"
#include "HttLkup.h"
extern htt_t			*htt;

static HttIcidLinkRec		*free_icid_link = 0;

static HttFocusWindowRec	*free_focuswindow_link = 0;
static HttFocusWindowRec	*busy_focuswindow_link = 0;

static void	set_free_ic_list();
static void	set_ic_list();
static void	del_client_rec();
static void	clear_client_associated_data();
static void	add_ic_list();

int
InitialClientTable()
{
    set_free_ic_list(0);
    set_ic_list(0);

    return (0);
}

static void
set_free_ic_list(ic)
   ic_t		*ic;
{
   htt->free_ic_list = ic;
}

static void
set_ic_list(ic)
   ic_t		*ic;
{
   htt->ic_list = ic;
}

static HttIcidLinkRec *
create_icid_link_rec()
{
    HttIcidLinkRec	*icid_link;

#ifdef DEBUG
    dbg_printf( "%% create_icid_link_rec\n" );
#endif
    if(free_icid_link) {
	icid_link = free_icid_link;
	if(free_icid_link->next)
		free_icid_link->next->prev = 0;
	free_icid_link = free_icid_link->next;
    }
    else {
	icid_link = (HttIcidLinkRec *)malloc(sizeof(HttIcidLinkRec));
	if(!icid_link)
		return((HttIcidLinkRec *)NULL);
	bzero(icid_link, sizeof(HttIcidLinkRec));
    }	
    return(icid_link);
}

static void
destroy_icid_link_rec(icid_link, focus_window_rec)
    HttIcidLinkRec	*icid_link;
    HttFocusWindowRec	*focus_window_rec;
{

#ifdef DEBUG
    dbg_printf( "%% destroy_icid_link_rec icid_link=%x\n", icid_link );
#endif
    if(icid_link->prev) 
	icid_link->prev->next = icid_link->next;
    else {
	if(focus_window_rec)
		focus_window_rec->icid_link = icid_link->next;
    }
    if(icid_link->next)
	icid_link->next->prev = icid_link->prev;

    icid_link->next = free_icid_link;
    if(free_icid_link)
	free_icid_link->prev = icid_link;
    free_icid_link = icid_link;
    icid_link->prev = 0;
    icid_link->icid = 0;
    return;
}

static HttFocusWindowRec *
get_focuswindow_rec(focus_window)
    Window		focus_window;
{
    HttFocusWindowRec	*window_rec;

#ifdef DEBUG
    dbg_printf( "%% get_focuswindow_rec\n" );
#endif
    for(window_rec = busy_focuswindow_link; window_rec; 
					window_rec = window_rec->next){
	if(window_rec->focus_window ==  focus_window)
		return (window_rec);
    }
    
    return (HttFocusWindowRec *)0;
}

static HttFocusWindowRec *
create_focuswindow_rec()
{
    HttFocusWindowRec	*window_rec;

#ifdef DEBUG
    dbg_printf( "%% create_focuswindow_rec\n" );
#endif
    if(free_focuswindow_link) {
	window_rec = free_focuswindow_link;
	if(free_focuswindow_link->next)
		free_focuswindow_link->next->prev = 0;
	free_focuswindow_link = free_focuswindow_link->next;
	bzero(window_rec, sizeof(HttFocusWindowRec));
    }
    else {
	window_rec = (HttFocusWindowRec *)malloc(sizeof(HttFocusWindowRec));
	if(!window_rec)
		return (HttFocusWindowRec *)0;
	bzero(window_rec, sizeof(HttFocusWindowRec));
    }
    window_rec->next = busy_focuswindow_link;
    if(busy_focuswindow_link)
	busy_focuswindow_link->prev = window_rec;
    busy_focuswindow_link = window_rec;
    return (window_rec);
}

static void
set_icid_link_rec(focus_window_rec, icid_link)
    HttFocusWindowRec	*focus_window_rec;
    HttIcidLinkRec	*icid_link;
{
    HttIcidLinkRec	*last_icid_link;

#ifdef DEBUG
    dbg_printf( "%% set_icid_link_rec\n" );
#endif
    focus_window_rec->count++;
    if ( !focus_window_rec->icid_link ) {
	focus_window_rec->icid_link = icid_link;
	return;
    }

    icid_link->next = focus_window_rec->icid_link;
    icid_link->prev = 0;
    if(icid_link->next)
	icid_link->next->prev = icid_link;
    focus_window_rec->icid_link = icid_link;
}

static void
release_icid_link_rec(focus_window_rec, icid)
    HttFocusWindowRec	*focus_window_rec;
    FXimp_Client	*icid;
{
    HttIcidLinkRec	*icid_link;

#ifdef DEBUG
    dbg_printf( "%% release_icid_link_rec  win_rec=%x, icid=%x\n",
			focus_window_rec, icid );
#endif
    if ( !focus_window_rec )
	return;

    icid_link = focus_window_rec->icid_link;
    for ( ; icid_link; icid_link = icid_link->next ) {
	if ( icid_link->icid != icid )
	    continue;

	destroy_icid_link_rec(icid_link, focus_window_rec);
	if ( focus_window_rec->current_icid == icid )
	    focus_window_rec->current_icid = ( focus_window_rec->icid_link ?
						icid : 0 );
	focus_window_rec->count--;
	if ( focus_window_rec->count <= 0 ) {
	    focus_window_rec->focus_window = 0;
	    focus_window_rec->count = 0;
	    focus_window_rec->icid_link = 0;
	    focus_window_rec->current_icid = 0;

	    if(focus_window_rec == busy_focuswindow_link) {
		busy_focuswindow_link = focus_window_rec->next;
		focus_window_rec->next->prev = 0;
	    }
	    if(focus_window_rec->prev)
		focus_window_rec->prev->next = focus_window_rec->next;
	    if(focus_window_rec->next)
		focus_window_rec->next->prev = focus_window_rec->prev;
	
	    if(free_focuswindow_link)
		free_focuswindow_link->prev = focus_window_rec;
	    focus_window_rec->next = free_focuswindow_link;
	    free_focuswindow_link  = focus_window_rec;
	}
	return;
    }
    return;
}

int
ExitClientTable()
{
    ic_t		*ic, *icn;
    FXimp_Client	*client;
    HttIcidLinkRec	*icid, *icidn;
    HttFocusWindowRec	*win_rec, *win_recn;

    ic = htt->ic_list;
    while(ic) {
	client = CORE_IC_REC(ic);
	del_client_rec(client);
	icn = ic->next;
	free(ic);
	ic = icn;
    }

    ic = htt->free_ic_list;
    while(ic) {
	client = CORE_IC_REC(ic);
	del_client_rec(client);
	icn = ic->next;
	free(ic);
	ic = icn;
    }

    icid = free_icid_link;
    while(icid) {
	icidn = icid->next;
	free(icid);
	icid = icidn;
    }

    win_rec = free_focuswindow_link;
    while(win_rec) {
	win_recn = win_rec->next;
	free(win_rec);
	win_rec = win_recn;
    }

    win_rec = busy_focuswindow_link;
    while(win_rec) {
	win_recn = win_rec->next;
	free(win_rec);
	win_rec = win_recn;
    }
    return (0);
}

static void
del_client_rec(client)
    FXimp_Client   *client;
{
    free(client->preedit);
    free(client->status);
    free(client->ims_area);
    free(client->ims_preedit);
    free(client->preedit_font);
    free(client->status_font);
    free(client);
}

FXimp_Client   *
NewClient(window)
    Window          window;
{
    Display     *display = GET_HTT_VALUE(htt, display);
    FXimp_Client	*client, *root;
    XWindowAttributes   win_att;
    ic_t		*ic;

    XGetWindowAttributes(display, window, &win_att);

    if(ic = htt->free_ic_list) {
	Ximp_PreeditPropRec 	*preedit_sav;
	Ximp_StatusPropRec	*status_sav;
	Ximp_AreaRec		*ims_area_sav;
	Ximp_AreaRec		*ims_preedit_sav;
	FXimpFont		*preedit_font_sav;
	FXimpFont		*status_font_sav;

	if(ic->next)
		ic->next->prev = 0;
	set_free_ic_list(ic->next);
	client = CORE_IC_REC(ic);

	preedit_sav     = client->preedit;
	status_sav	= client->status;
	ims_area_sav	= client->ims_area;
	ims_preedit_sav = client->ims_preedit;
	preedit_font_sav= client->preedit_font;
	status_font_sav = client->status_font;

	bzero(client, sizeof(FXimp_Client));
	
	client->preedit      = preedit_sav;
	client->status	     = status_sav;
	client->ims_area     = ims_area_sav;
	client->ims_preedit  = ims_preedit_sav;
	client->preedit_font = preedit_font_sav;
	client->status_font  = status_font_sav;
	clear_client_associated_data(client);
	add_ic_list(ic);
    }
    else {
	ic = (ic_t *)malloc(sizeof(ic_t));
	if(!ic)
		return (FXimp_Client *)NULL;
	client = (FXimp_Client *)malloc(sizeof(FXimp_Client));
	if(!client) {
		free(ic);
		return (FXimp_Client *)NULL;
	}
	bzero(client, sizeof(FXimp_Client));
	ic->core_ic = client;
	client->preedit = 
		(Ximp_PreeditPropRec *)malloc(sizeof(Ximp_PreeditPropRec));
	client->status  = 
		(Ximp_StatusPropRec *)malloc(sizeof(Ximp_StatusPropRec));
	client->ims_area =  (Ximp_AreaRec *)malloc(sizeof(Ximp_AreaRec));
	client->ims_preedit = (Ximp_AreaRec *)malloc(sizeof(Ximp_AreaRec));
	client->preedit_font = (FXimpFont *)malloc(sizeof(FXimpFont));
	client->status_font = (FXimpFont *)malloc(sizeof(FXimpFont));
	clear_client_associated_data(client);
	add_ic_list(ic);
    }
    
    client->window = window;
    client->ims = htt;
    if(window != 1)
    	client->screen = XScreenNumberOfScreen(win_att.screen);

    return (client);
          
}

static void
clear_client_associated_data(client)
    FXimp_Client	*client;
{
    bzero(client->preedit,     sizeof(Ximp_PreeditPropRec));
    bzero(client->status,      sizeof(Ximp_StatusPropRec));
    bzero(client->ims_area,    sizeof(Ximp_AreaRec));
    bzero(client->ims_preedit, sizeof(Ximp_AreaRec));
    bzero(client->preedit_font,sizeof(FXimpFont));
    bzero(client->status_font, sizeof(FXimpFont));
}

static void
add_ic_list(ic)
    ic_t	*ic;
{
    ic->next = htt->ic_list;
    if(ic->next)
	ic->next->prev = ic;
    htt->ic_list = ic;
    ic->prev = 0;
}


int
DeleteClient(icid)
    FXimp_ICID      icid;
{
    FXimp_Client   *root_client = GET_HTT_VALUE(htt, root_client);
    ic_t	*ic;

    if(!htt->ic_list)
	return (-1);

    if (icid == (FXimp_ICID) CURRENT_CLIENT(htt)) {
	SET_HTT_VALUE(htt, current_client, root_client);
    }

    for(ic = htt->ic_list; ic ; ic = ic->next) {
	if(CORE_IC_REC(ic) == icid) {
		if(ic->prev)
			ic->prev->next = ic->next;
		else
			htt->ic_list = ic->next;
		if(ic->next)
			ic->next->prev = ic->prev;

		ic->prev = 0;
		ic->next = htt->free_ic_list;
		if(htt->free_ic_list)
			(htt->free_ic_list)->prev = ic;
		htt->free_ic_list = ic;
		return(0);
	}
    }

    return(-1);
}

FXimp_Client   *
FindClientByICID(icid)
    FXimp_ICID      icid;
{
    return ((FXimp_Client *)icid);
}

FXimp_Client   *
FindClient(window)
    Window          window;
{
    ic_t        *root_ic, *tmp;
    FXimp_Client	*client;

    root_ic = htt->ic_list;

    for(tmp=root_ic; tmp; tmp=tmp->next) {
	client = CORE_IC_REC(tmp);
	if(client->window == window)
		return (client);
    }

    return ((FXimp_Client *) 0);/* error return */
}

static HttFocusWindowRec   *
find_focus_window_rec(window)
    Window		 window;
{
    int			 i;
    HttFocusWindowRec	*window_rec;

#ifdef DEBUG
    dbg_printf( "%% find_focus_window_rec  focus_window=%x\n", window );
#endif
    for(window_rec = busy_focuswindow_link; window_rec; 
					window_rec = window_rec->next){
	if(window_rec->focus_window == window) {
		return(window_rec);
	}
    }
    return (HttFocusWindowRec *)0;
}

FXimp_Client   *
FindClientByFocusWindow(window)
    Window		 window;
{
    int			 i;
    HttFocusWindowRec	*focus_window_rec;

    focus_window_rec = find_focus_window_rec( window );
    if ( !focus_window_rec )
	return (FXimp_Client *)0;
    return (FXimp_Client *)focus_window_rec->current_icid;
}

FXimp_Client   *
FindClientByPreeditOrStatusWindow(window)
    Window          window;
{
    ic_t        *root_ic, *tmp;
    FXimp_Client	*client;

    root_ic = htt->ic_list;

    for(tmp = root_ic; tmp ; tmp = tmp->next) {
	client = CORE_IC_REC(tmp);
	if(client->preedit_window == window ||
		client->status_window == window ) {
		return (client);
	}
    }

    return ((FXimp_Client *) 0);/* error return */
}

FXimp_Client    *
FindClientByPreeditOrStatusWindowForRedisplay(window)
Window  window;
{
    ic_t        *root_ic, *tmp;
    FXimp_Client        *client;
    int preedit_feedback, status_feedback;

    root_ic = htt->ic_list;


    for(tmp = root_ic; tmp ; tmp = tmp->next) {
        client = CORE_IC_REC(tmp);
        if(client->preedit_window == window ||
                client->status_window == window ) {
            if ( ((client->preedit_window == window) &&
                  (client->style & XIMPreeditNothing)) ||
                 ((client->status_window == window) &&
                  (client->style & XIMStatusNothing)) ) {
                if (client == GET_HTT_VALUE(htt,previous_focus_in_client))
                        return(client);
             }
             else
                return(client);
        }
    }
 
    return ((FXimp_Client *) 0);/* error return */
}
 
int
SetCurrentClient(window)
    Window          window;
{
    FXimp_Client   *root_client = GET_HTT_VALUE(htt, root_client);
    FXimp_Client   *ximp_client;

    ximp_client = FindClientByFocusWindow(window);
    if (!ximp_client) {
	ximp_client = FindClient(window);
	if (!ximp_client) {
	    SET_HTT_VALUE(htt, current_client, root_client);
	    return (-1);
	}
    }
    SET_HTT_VALUE(htt, current_client, ximp_client);
    return (0);
}

DestroyCurrentClient(event)
    XEvent         *event;
{
    Display        *display = GET_HTT_VALUE(htt, display);
    FXimp_Client   *root_client = GET_HTT_VALUE(htt, root_client);

    if (CURRENT_CLIENT(htt)->status_window ==
        root_client->status_window)
        unmap_status_window(CURRENT_CLIENT(htt));
    if (CURRENT_CLIENT(htt)->preedit_window ==
        root_client->preedit_window)
        unmap_preedit_window(CURRENT_CLIENT(htt));

    if (CURRENT_CLIENT(htt) == GET_HTT_VALUE(htt, previous_focus_in_client))
        SET_HTT_VALUE(htt, previous_focus_in_client, NULL);

    if (CURRENT_CLIENT(htt) == GET_HTT_VALUE(htt, previous_client))
	SET_HTT_VALUE(htt, previous_client, NULL);

    if (CURRENT_CLIENT(htt)->prot_type == FPRT_XIM) {
	CURRENT_CLIENT(htt)->focus_window = 0;
	XimpDestroyClient(CURRENT_CLIENT(htt), 1);
    }
#ifdef XJP
    else {
	CURRENT_CLIENT(htt)->focus_window = 0;
	XjpDestroyClient(CURRENT_CLIENT(htt), 1);
    }
#endif	/* XJP */
}

EntryFocusWindow( icid, focus_window )
    FXimp_Client	*icid;
    Window		 focus_window;
{
    HttFocusWindowRec	*focus_window_rec;
    HttIcidLinkRec	*icid_link;

#ifdef DEBUG
    dbg_printf( "%% EntryFocusWindow\n" );
#endif
    if ( ( icid_link = create_icid_link_rec() )
					== (HttIcidLinkRec *)0 )
	return -1;
    icid_link->icid = (struct _FXimp_Client *)icid;
    if ( ( focus_window_rec = get_focuswindow_rec(focus_window) )
					== (HttFocusWindowRec *)0 ) {
	if ( ( focus_window_rec = create_focuswindow_rec() )
					== (HttFocusWindowRec *)0 ) {
	    destroy_icid_link_rec(icid_link, focus_window_rec);
	    return -1;
	}
	focus_window_rec->focus_window = focus_window;
    }
    set_icid_link_rec( focus_window_rec, icid_link );
    focus_window_rec->current_icid = (struct _FXimp_Client *)icid;
    icid->focus_window_rec = focus_window_rec;
    return 0;
}

DestroyClientByFocus( window )
    Window		 window;
{
    HttFocusWindowRec	*focus_window_rec;
    HttIcidLinkRec	*icid_link;

#ifdef DEBUG
    dbg_printf( "%% DestroyFocusWindow\n" );
#endif
    focus_window_rec = find_focus_window_rec( window );
    if ( !focus_window_rec ) {
	if ( SetCurrentClient( window ) == 0 ) {
	    DestroyCurrentClient(0);
	}
	return 0;
    }

    for ( icid_link = focus_window_rec->icid_link; icid_link; ) {
	SET_HTT_VALUE(htt, current_client, (FXimp_Client *)icid_link->icid);
	icid_link = icid_link->next;
	DestroyCurrentClient(0);
    }
    return;
}

SetCurrentClientInFocus( icid )
    FXimp_Client	*icid;
{
#ifdef DEBUG
    dbg_printf( "%% SetCurrentClientInFocus  icid=%x\n", icid );
#endif
    icid->focus_window_rec->current_icid = (struct _FXimp_Client *)icid;
    return 0;
}

DeleteFocusWindow( icid )
    FXimp_Client	*icid;
{
#ifdef DEBUG
    dbg_printf( "%% DeleteFocusWindow\n" );
#endif
    if ( icid->focus_window_rec ) 
	release_icid_link_rec( icid->focus_window_rec, icid );
    return 0;
}

NumOfRemainConvOnIC(current_client)
    FXimp_Client   *current_client;
{
    HttIcidLinkRec	*sibling;
    int			count = 0;

    sibling = current_client->focus_window_rec->icid_link;

    for( ; sibling; sibling = sibling->next){
	if(sibling->icid->imp_mode & FIMP_BEGIN)
		count++;
    }
    return(count);
}

int
CheckRootWindow()
{
    Display             *display = GET_HTT_VALUE(htt, display);
    FXimp_Client        *root_client = GET_HTT_VALUE(htt, root_client);
    ic_t                *root_ic, *tmp;
    FXimp_Client        *client;
    int                  num = 0;
    LookupInfo *info =  (LookupInfo *)GET_HTT_VALUE(htt, lookup_info);

        if(info->style == LOOKUP_MENU_IMSWINDOW ||
           info->style == LOOKUP_MENU_IMSWINDOW2) {
               return (1);
        }

        root_ic = htt->ic_list;
        for(tmp = root_ic; tmp ; tmp = tmp->next) {
            client = CORE_IC_REC(tmp);
            if (client->style & XIMPreeditNothing ||
                client->style & XIMStatusNothing){
                    if(tmp->next) {
                        return (1);
                    }
            }
            num++;
         }
     XUnmapWindow(display, root_client->window);
     return (0);
}

FXimp_Client	*
FindAnyActiveClient(head)
    Bool	head;
{
    static HttFocusWindowRec *fwinrec;
    static HttIcidLinkRec    *iclink;
    HttIcidLinkRec	     *tmp;
    Bool		     new_iclink;

    if(head) {
	fwinrec = busy_focuswindow_link;
	new_iclink = True;
    }
    else
	new_iclink = False;
    
    for(; fwinrec; fwinrec = fwinrec->next) {
	if(new_iclink) {
		iclink = fwinrec->icid_link; 
		new_iclink = False;
	}	
	for(; iclink; iclink = iclink->next) {
		if(iclink->icid && iclink->icid->s) {
			tmp = iclink;
			iclink = iclink->next;
			return (tmp->icid);
		}
	}
	new_iclink = True;
    }
    return (FXimp_Client *)NULL;
}
