/******************************************************************

       Copyright 1990, 1991, 1992, 1993 by Sun Microsystems, Inc.

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and
that both that copyright notice and this permission notice appear
in supporting documentation, and that the name of Sun Microsystems, Inc.
not be used in advertising or publicity pertaining to distribution
of the software without specific, written prior permission.
Sun Microsystems, Inc. makes no representations about the suitability of
this software for any purpose.  It is provided "as is" without
express or implied warranty.

Sun Microsystems Inc. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
IN NO EVENT SHALL Sun Microsystems, Inc. BE LIABLE FOR ANY SPECIAL, INDIRECT
OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
OR PERFORMANCE OF THIS SOFTWARE.

  Author: Hideki Hiura (hhiura@Sun.COM)
	  				     Sun Microsystems, Inc.
******************************************************************/

#ifndef lint
#ident  "@(#)XSunIMConf.c	1.20    93/02/12 SMI"
#endif

#include <stdio.h>
#include <dlfcn.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include "XSunExt.h"
#include <X11/keysym.h>
#include "XSunIMProt.h"
#include "XSunIMMMan.h"
#include "XSunIMMthd.h"
#include "XSunIMPriv.h"
#include "XSunIMCore.h"
#include "XSunIMPub.h"

#ifdef SVR4
#include <libintl.h>
#endif /* SVR4 */


typedef struct _iml_le_sw {
    char *if_name ;
    if_methods_t *m ;
} iml_le_sw_t ;

#ifdef BuiltInCm
extern if_methods_t cm_methods;
#endif
#ifdef BuiltInWnn
extern if_methods_t wnn4_methods;
#endif
#ifdef BuiltInXci
extern if_methods_t xci_methods;
#endif
#ifdef BuiltInSj3
extern if_methods_t sj3_methods;
#endif
#ifdef BuiltInCanna
extern if_methods_t canna_methods;
#endif

Private iml_le_sw_t iml_le_sw[]={
#ifdef BuiltInCm
    {"cm", &cm_methods },
#endif
#ifdef BuiltInWnn
    {"wnn4", &wnn4_methods },
#endif
#ifdef BuiltInXci
    {"xci", &xci_methods },
#endif
#ifdef BuiltInSj3
    {"sj3", &sj3_methods },
#endif
#ifdef BuiltInCanna
    {"canna", &canna_methods },
#endif
    {"", NULL}
};
#define HOW_MANY_LES_STATICALLY_ATTACHED   ((sizeof(iml_le_sw)-1) / sizeof(iml_le_sw_t))

#define IFPATH "HTT_LIBRARY_PATH"
#ifdef _POSIX_PATH_MAX
#define PATHNAME_MAX _POSIX_PATH_MAX
#else
#define PATHNAME_MAX 1024
#endif
#define SYMNAME_MAX 255

Private char *ow_home_default="/usr/openwin" ;
Public iml_methods_t *_iml_dup_iml_methods();
Public char *getenv();

/*
 * if_configure $@$O!"(J1 interface $@$K$D$-!"(Jlocale $@$4$H$K(J 1 $@2s$N$_8F$P$l$k!#(J
 */
Public Bool
if_configure(If, if_name, locale, im)
iml_if_t *If ;
char *if_name ;
char *locale ;
void  *im ;
{
    static char *ow_home ;
    char *if_dir = getenv(IFPATH);
    char if_path[PATHNAME_MAX] ;
    char if_entry[SYMNAME_MAX] ;
    void *if_so ;
    register i ;
    int is_statically_attached = 0 ;

    for (i=0 ; iml_le_sw[i].m ; i++){
	if(!strcmp(iml_le_sw[i].if_name, if_name)){
	    If->ifm = iml_le_sw[i].m ;
	    goto Found;
	}
    }
#ifndef DYNAMICIF
    /*
     * Fall down to search dynamic if
     */
    if(!if_dir){
	if(!(ow_home = getenv("OPENWINHOME"))){
	    ow_home = ow_home_default ;
	}
	sprintf(if_path, "%s/lib/locale/%s/htt/%s.so", ow_home, locale, if_name);
    } else {
	sprintf(if_path, "%s/%s/htt/%s.so", if_dir, locale, if_name);
    }
    if_so = dlopen(if_path, RTLD_LAZY);
    if(if_so){
	sprintf(if_entry, "%s_methods", if_name);
	If->ifm = dlsym(if_so, if_entry);
    }
#endif /* ifndef DYNAMICIF */
    if(!if_so || !If->ifm ){
#ifdef SVR4
	fprintf(stderr, dgettext("SUNW_XIM_HTT_MSG","htt: Engine I/F loading - %s\n"), dlerror);
	fprintf(stderr,
          dgettext("SUNW_XIM_HTT_MSG","htt: if_configure error - Language Engine %s is not pre-attached\n"),
          if_name);
#else /* SVR4 */
        fprintf(stderr, "htt: Engine I/F loading - %s\n",dlerror);
        fprintf(stderr, "htt: if_configure error - Language Engine %s is not pre-attached\n",if_name);
#endif /* SVR4 */
	return False ;
    }
Found:    
    If->m = _iml_dup_iml_methods();
    if(If->ifm->if_OpenIF(If, locale, im) == False){
	return False ;
    }
    if(If->ifm->if_GetIFValue(If, IF_VERSION, &If->if_version) == True){
/*
 * Following messages are hibernated.
 */
#ifdef SVR4
/*	printf(dgettext("SUNW_XIM_HTT_MSG","htt: Language Engine I/F \"%s\" version %s for locale %s is attached.\n"), */
#else /* SVR4 */ 
/*        printf("htt: Language Engine I/F \"%s\" version %s for locale %s is attached.\n", */
#endif /* SVR4 */
#ifdef DEBUG
#ifdef DYNAMICIF
/*	       if_path, */
#else
/*	       if_name, */
#endif
#else
/*	       if_name, */
#endif	
/*	       If->if_version, locale); */
    } else {
#ifdef SVR4
/*	printf(dgettext("SUNW_XIM_HTT_MSG","htt: Language Engine I/F %s for the locale %s is attached\n"), */
#else /* SVR4 */
/*        printf("htt: Language Engine I/F %s for the locale %s is attached\n", */
#endif /* SVR4 */
#ifdef DEBUG
#ifdef DYNAMICIF
/*	       if_path, */
#else
/*	       if_name, */
#endif
#else
/*	       if_name, */
#endif	
/*	       locale); */
    }
    return True ;
}	

